@@ -1,10 +1,48 @@
+#  Licensed to Elasticsearch B.V. under one or more contributor
+#  license agreements. See the NOTICE file distributed with
+#  this work for additional information regarding copyright
+#  ownership. Elasticsearch B.V. licenses this file to you under
+#  the Apache License, Version 2.0 (the "License"); you may
+#  not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+# 	http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing,
+#  software distributed under the License is distributed on an
+#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#  KIND, either express or implied.  See the License for the
+#  specific language governing permissions and limitations
+#  under the License.
+
+#  Licensed to Elasticsearch B.V. under one or more contributor
+#  license agreements. See the NOTICE file distributed with
+#  this work for additional information regarding copyright
+#  ownership. Elasticsearch B.V. licenses this file to you under
+#  the Apache License, Version 2.0 (the "License"); you may
+#  not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+# 	http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing,
+#  software distributed under the License is distributed on an
+#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#  KIND, either express or implied.  See the License for the
+#  specific language governing permissions and limitations
+#  under the License.
+import warnings
+from datetime import datetime, timedelta
+
+import pytest
+from dateutil import tz
 from mock import patch
 
-from elasticsearch import helpers, TransportError
+from elasticsearch import TransportError, helpers
 from elasticsearch.helpers import ScanError
 
-from . import ElasticsearchTestCase
 from ..test_cases import SkipTest
+from . import ElasticsearchTestCase
 
 
 class FailingBulkClient(object):
@@ -31,17 +69,19 @@ class TestStreamingBulk(ElasticsearchTestCase):
             self.client, actions, index="test-index"
         ):
             self.assertTrue(ok)
-        self.assertEquals([{"_id": 1}, {"_id": 2}], actions)
+        self.assertEqual([{"_id": 1}, {"_id": 2}], actions)
 
     def test_all_documents_get_inserted(self):
         docs = [{"answer": x, "_id": x} for x in range(100)]
-        for ok, item in helpers.streaming_bulk(
-            self.client, docs, index="test-index", refresh=True
-        ):
-            self.assertTrue(ok)
+        with warnings.catch_warnings(record=True) as w:
+            for ok, item in helpers.streaming_bulk(
+                self.client, docs, index="test-index", refresh=True
+            ):
+                self.assertTrue(ok)
 
-        self.assertEquals(100, self.client.count(index="test-index")["count"])
-        self.assertEquals(
+        self.assertEqual(w, [])
+        self.assertEqual(100, self.client.count(index="test-index")["count"])
+        self.assertEqual(
             {"answer": 42}, self.client.get(index="test-index", id=42)["_source"]
         )
 
@@ -61,12 +101,12 @@ class TestStreamingBulk(ElasticsearchTestCase):
             ):
                 self.assertTrue(ok)
         except helpers.BulkIndexError as e:
-            self.assertEquals(2, len(e.errors))
+            self.assertEqual(2, len(e.errors))
         else:
             assert False, "exception should have been raised"
 
     def test_different_op_types(self):
-        if self.es_version < (0, 90, 1):
+        if self.es_version() < (0, 90, 1):
             raise SkipTest("update supported since 0.90.1")
         self.client.index(index="i", id=45, body={})
         self.client.index(index="i", id=42, body={})
@@ -85,8 +125,8 @@ class TestStreamingBulk(ElasticsearchTestCase):
             self.assertTrue(ok)
 
         self.assertFalse(self.client.exists(index="i", id=45))
-        self.assertEquals({"answer": 42}, self.client.get(index="i", id=42)["_source"])
-        self.assertEquals({"f": "v"}, self.client.get(index="i", id=47)["_source"])
+        self.assertEqual({"answer": 42}, self.client.get(index="i", id=42)["_source"])
+        self.assertEqual({"f": "v"}, self.client.get(index="i", id=47)["_source"])
 
     def test_transport_error_can_becaught(self):
         failing_client = FailingBulkClient(self.client)
@@ -105,13 +145,13 @@ class TestStreamingBulk(ElasticsearchTestCase):
                 chunk_size=1,
             )
         )
-        self.assertEquals(3, len(results))
-        self.assertEquals([True, False, True], [r[0] for r in results])
+        self.assertEqual(3, len(results))
+        self.assertEqual([True, False, True], [r[0] for r in results])
 
         exc = results[1][1]["index"].pop("exception")
         self.assertIsInstance(exc, TransportError)
-        self.assertEquals(599, exc.status_code)
-        self.assertEquals(
+        self.assertEqual(599, exc.status_code)
+        self.assertEqual(
             {
                 "index": {
                     "_index": "i",
@@ -145,12 +185,12 @@ class TestStreamingBulk(ElasticsearchTestCase):
                 initial_backoff=0,
             )
         )
-        self.assertEquals(3, len(results))
-        self.assertEquals([True, True, True], [r[0] for r in results])
+        self.assertEqual(3, len(results))
+        self.assertEqual([True, True, True], [r[0] for r in results])
         self.client.indices.refresh(index="i")
         res = self.client.search(index="i")
-        self.assertEquals({"value": 3, "relation": "eq"}, res["hits"]["total"])
-        self.assertEquals(4, failing_client._called)
+        self.assertEqual({"value": 3, "relation": "eq"}, res["hits"]["total"])
+        self.assertEqual(4, failing_client._called)
 
     def test_rejected_documents_are_retried_at_most_max_retries_times(self):
         failing_client = FailingBulkClient(
@@ -173,12 +213,12 @@ class TestStreamingBulk(ElasticsearchTestCase):
                 initial_backoff=0,
             )
         )
-        self.assertEquals(3, len(results))
-        self.assertEquals([False, True, True], [r[0] for r in results])
+        self.assertEqual(3, len(results))
+        self.assertEqual([False, True, True], [r[0] for r in results])
         self.client.indices.refresh(index="i")
         res = self.client.search(index="i")
-        self.assertEquals({"value": 2, "relation": "eq"}, res["hits"]["total"])
-        self.assertEquals(4, failing_client._called)
+        self.assertEqual({"value": 2, "relation": "eq"}, res["hits"]["total"])
+        self.assertEqual(4, failing_client._called)
 
     def test_transport_error_is_raised_with_max_retries(self):
         failing_client = FailingBulkClient(
@@ -200,7 +240,7 @@ class TestStreamingBulk(ElasticsearchTestCase):
             return results
 
         self.assertRaises(TransportError, streaming_bulk)
-        self.assertEquals(4, failing_client._called)
+        self.assertEqual(4, failing_client._called)
 
 
 class TestBulk(ElasticsearchTestCase):
@@ -210,10 +250,10 @@ class TestBulk(ElasticsearchTestCase):
             self.client, docs, index="test-index", refresh=True
         )
 
-        self.assertEquals(1, success)
+        self.assertEqual(1, success)
         self.assertFalse(failed)
-        self.assertEquals(1, self.client.count(index="test-index")["count"])
-        self.assertEquals(
+        self.assertEqual(1, self.client.count(index="test-index")["count"])
+        self.assertEqual(
             {"answer": 42}, self.client.get(index="test-index", id=1)["_source"]
         )
 
@@ -223,10 +263,10 @@ class TestBulk(ElasticsearchTestCase):
             self.client, docs, index="test-index", refresh=True
         )
 
-        self.assertEquals(100, success)
+        self.assertEqual(100, success)
         self.assertFalse(failed)
-        self.assertEquals(100, self.client.count(index="test-index")["count"])
-        self.assertEquals(
+        self.assertEqual(100, self.client.count(index="test-index")["count"])
+        self.assertEqual(
             {"answer": 42}, self.client.get(index="test-index", id=42)["_source"]
         )
 
@@ -236,9 +276,9 @@ class TestBulk(ElasticsearchTestCase):
             self.client, docs, index="test-index", refresh=True, stats_only=True
         )
 
-        self.assertEquals(100, success)
-        self.assertEquals(0, failed)
-        self.assertEquals(100, self.client.count(index="test-index")["count"])
+        self.assertEqual(100, success)
+        self.assertEqual(0, failed)
+        self.assertEqual(100, self.client.count(index="test-index")["count"])
 
     def test_errors_are_reported_correctly(self):
         self.client.indices.create(
@@ -256,12 +296,12 @@ class TestBulk(ElasticsearchTestCase):
             index="i",
             raise_on_error=False,
         )
-        self.assertEquals(1, success)
-        self.assertEquals(1, len(failed))
+        self.assertEqual(1, success)
+        self.assertEqual(1, len(failed))
         error = failed[0]
-        self.assertEquals("42", error["index"]["_id"])
-        self.assertEquals("_doc", error["index"]["_type"])
-        self.assertEquals("i", error["index"]["_index"])
+        self.assertEqual("42", error["index"]["_id"])
+        self.assertEqual("_doc", error["index"]["_type"])
+        self.assertEqual("i", error["index"]["_index"])
         print(error["index"]["error"])
         self.assertTrue(
             "MapperParsingException" in repr(error["index"]["error"])
@@ -286,6 +326,39 @@ class TestBulk(ElasticsearchTestCase):
             index="i",
         )
 
+    def test_ignore_error_if_raised(self):
+        # ignore the status code 400 in tuple
+        helpers.bulk(
+            self.client, [{"a": 42}, {"a": "c"}], index="i", ignore_status=(400,)
+        )
+
+        # ignore the status code 400 in list
+        helpers.bulk(
+            self.client,
+            [{"a": 42}, {"a": "c"}],
+            index="i",
+            ignore_status=[
+                400,
+            ],
+        )
+
+        # ignore the status code 400
+        helpers.bulk(self.client, [{"a": 42}, {"a": "c"}], index="i", ignore_status=400)
+
+        # ignore only the status code in the `ignore_status` argument
+        self.assertRaises(
+            helpers.BulkIndexError,
+            helpers.bulk,
+            self.client,
+            [{"a": 42}, {"a": "c"}],
+            index="i",
+            ignore_status=(444,),
+        )
+
+        # ignore transport error exception
+        failing_client = FailingBulkClient(self.client)
+        helpers.bulk(failing_client, [{"a": 42}], index="i", ignore_status=(599,))
+
     def test_errors_are_collected_properly(self):
         self.client.indices.create(
             "i",
@@ -303,8 +376,8 @@ class TestBulk(ElasticsearchTestCase):
             stats_only=True,
             raise_on_error=False,
         )
-        self.assertEquals(1, success)
-        self.assertEquals(1, failed)
+        self.assertEqual(1, success)
+        self.assertEqual(1, failed)
 
 
 class TestScan(ElasticsearchTestCase):
@@ -321,10 +394,9 @@ class TestScan(ElasticsearchTestCase):
         },
     ]
 
-    @classmethod
-    def tearDownClass(cls):
-        cls.client.transport.perform_request("DELETE", "/_search/scroll/_all")
-        super(TestScan, cls).tearDownClass()
+    def teardown_method(self, m):
+        self.client.transport.perform_request("DELETE", "/_search/scroll/_all")
+        super(TestScan, self).teardown_method(m)
 
     def test_order_can_be_preserved(self):
         bulk = []
@@ -333,18 +405,22 @@ class TestScan(ElasticsearchTestCase):
             bulk.append({"answer": x, "correct": x == 42})
         self.client.bulk(bulk, refresh=True)
 
-        docs = list(
-            helpers.scan(
-                self.client,
-                index="test_index",
-                query={"sort": "answer"},
-                preserve_order=True,
+        with warnings.catch_warnings(record=True) as w:
+            docs = list(
+                helpers.scan(
+                    self.client,
+                    index="test_index",
+                    query={"sort": "answer"},
+                    preserve_order=True,
+                )
             )
-        )
 
-        self.assertEquals(100, len(docs))
-        self.assertEquals(list(map(str, range(100))), list(d["_id"] for d in docs))
-        self.assertEquals(list(range(100)), list(d["_source"]["answer"] for d in docs))
+        # Asserts that no warnings are raised.
+        self.assertEqual(w, [])
+
+        self.assertEqual(100, len(docs))
+        self.assertEqual(list(map(str, range(100))), list(d["_id"] for d in docs))
+        self.assertEqual(list(range(100)), list(d["_source"]["answer"] for d in docs))
 
     def test_all_documents_are_read(self):
         bulk = []
@@ -355,9 +431,9 @@ class TestScan(ElasticsearchTestCase):
 
         docs = list(helpers.scan(self.client, index="test_index", size=2))
 
-        self.assertEquals(100, len(docs))
-        self.assertEquals(set(map(str, range(100))), set(d["_id"] for d in docs))
-        self.assertEquals(set(range(100)), set(d["_source"]["answer"] for d in docs))
+        self.assertEqual(100, len(docs))
+        self.assertEqual(set(map(str, range(100))), set(d["_id"] for d in docs))
+        self.assertEqual(set(range(100)), set(d["_source"]["answer"] for d in docs))
 
     def test_scroll_error(self):
         bulk = []
@@ -429,6 +505,120 @@ class TestScan(ElasticsearchTestCase):
             client_mock.scroll.assert_not_called()
             client_mock.clear_scroll.assert_not_called()
 
+    def test_scan_auth_kwargs_forwarded(self):
+        for key, val in {
+            "api_key": ("name", "value"),
+            "http_auth": ("username", "password"),
+            "headers": {"custom": "header"},
+        }.items():
+            with patch.object(self, "client") as client_mock:
+                client_mock.search.return_value = {
+                    "_scroll_id": "scroll_id",
+                    "_shards": {"successful": 5, "total": 5, "skipped": 0},
+                    "hits": {"hits": [{"search_data": 1}]},
+                }
+                client_mock.scroll.return_value = {
+                    "_scroll_id": "scroll_id",
+                    "_shards": {"successful": 5, "total": 5, "skipped": 0},
+                    "hits": {"hits": []},
+                }
+                client_mock.clear_scroll.return_value = {}
+
+                data = list(helpers.scan(self.client, index="test_index", **{key: val}))
+
+                self.assertEqual(data, [{"search_data": 1}])
+
+                # Assert that 'search', 'scroll' and 'clear_scroll' all
+                # received the extra kwarg related to authentication.
+                for api_mock in (
+                    client_mock.search,
+                    client_mock.scroll,
+                    client_mock.clear_scroll,
+                ):
+                    self.assertEqual(api_mock.call_args[1][key], val)
+
+    def test_scan_auth_kwargs_favor_scroll_kwargs_option(self):
+        with patch.object(self, "client") as client_mock:
+            client_mock.search.return_value = {
+                "_scroll_id": "scroll_id",
+                "_shards": {"successful": 5, "total": 5, "skipped": 0},
+                "hits": {"hits": [{"search_data": 1}]},
+            }
+            client_mock.scroll.return_value = {
+                "_scroll_id": "scroll_id",
+                "_shards": {"successful": 5, "total": 5, "skipped": 0},
+                "hits": {"hits": []},
+            }
+            client_mock.clear_scroll.return_value = {}
+
+            data = list(
+                helpers.scan(
+                    self.client,
+                    index="test_index",
+                    scroll_kwargs={"headers": {"scroll": "kwargs"}, "sort": "asc"},
+                    headers={"not scroll": "kwargs"},
+                )
+            )
+
+            self.assertEqual(data, [{"search_data": 1}])
+
+            # Assert that we see 'scroll_kwargs' options used instead of 'kwargs'
+            self.assertEqual(
+                client_mock.scroll.call_args[1]["headers"], {"scroll": "kwargs"}
+            )
+            self.assertEqual(client_mock.scroll.call_args[1]["sort"], "asc")
+
+    def test_scan_duplicate_parameters(self):
+        with patch.object(self.client, "search") as search_mock, patch.object(
+            self.client, "scroll"
+        ) as scroll_mock, patch.object(
+            self.client, "clear_scroll"
+        ) as clear_scroll_mock:
+            search_mock.return_value = {
+                "_scroll_id": "scroll_id",
+                "_shards": {"successful": 5, "total": 5, "skipped": 0},
+                "hits": {"hits": [{"field": "value"}]},
+            }
+
+            scroll_mock.return_value = {
+                "_scroll_id": "scroll_id",
+                "_shards": {"successful": 5, "total": 5, "skipped": 0},
+                "hits": {"hits": []},
+            }
+
+            clear_scroll_mock.return_value = {"acknowledged": True}
+            data = [
+                x
+                for x in helpers.scan(
+                    self.client,
+                    index="test_index",
+                    size=10,
+                    query={"size": 1},
+                    scroll_kwargs={"scroll": "10m", "rest_total_hits_as_int": True},
+                )
+            ]
+
+            assert data == [{"field": "value"}]
+            search_mock.assert_called_with(
+                index="test_index",
+                size=10,
+                sort="_doc",
+                scroll="5m",
+                request_timeout=None,
+                params={"__elastic_client_meta": (("h", "s"),)},
+            )
+            scroll_mock.assert_called_with(
+                scroll="5m",
+                rest_total_hits_as_int=True,
+                params={"__elastic_client_meta": (("h", "s"),)},
+                scroll_id="scroll_id",
+            )
+            clear_scroll_mock.assert_called_with(
+                scroll_id="scroll_id",
+                ignore=(404,),
+                params={"__elastic_client_meta": (("h", "s"),)},
+            )
+
     @patch("elasticsearch.helpers.actions.logger")
     def test_logger(self, logger_mock):
         bulk = []
@@ -492,10 +682,36 @@ class TestScan(ElasticsearchTestCase):
             )
             spy.assert_not_called()
 
+    def test_shards_no_skipped_field(self):
+        with patch.object(self, "client") as client_mock:
+            client_mock.search.return_value = {
+                "_scroll_id": "dummy_id",
+                "_shards": {"successful": 5, "total": 5},
+                "hits": {"hits": [{"search_data": 1}]},
+            }
+            client_mock.scroll.side_effect = [
+                {
+                    "_scroll_id": "dummy_id",
+                    "_shards": {"successful": 5, "total": 5},
+                    "hits": {"hits": [{"scroll_data": 42}]},
+                },
+                {
+                    "_scroll_id": "dummy_id",
+                    "_shards": {"successful": 5, "total": 5},
+                    "hits": {"hits": []},
+                },
+            ]
+
+            data = list(
+                helpers.scan(
+                    self.client, index="test_index", size=2, raise_on_error=True
+                )
+            )
+            self.assertEqual(data, [{"search_data": 1}, {"scroll_data": 42}])
+
 
 class TestReindex(ElasticsearchTestCase):
-    def setUp(self):
-        super(TestReindex, self).setUp()
+    def setup_method(self, _):
         bulk = []
         for x in range(100):
             bulk.append({"index": {"_index": "test_index", "_type": "_doc", "_id": x}})
@@ -518,11 +734,11 @@ class TestReindex(ElasticsearchTestCase):
         )
 
         self.assertTrue(self.client.indices.exists("prod_index"))
-        self.assertEquals(
+        self.assertEqual(
             50, self.client.count(index="prod_index", q="type:answers")["count"]
         )
 
-        self.assertEquals(
+        self.assertEqual(
             {"answer": 42, "correct": True, "type": "answers"},
             self.client.get(index="prod_index", id=42)["_source"],
         )
@@ -537,11 +753,11 @@ class TestReindex(ElasticsearchTestCase):
         self.client.indices.refresh()
 
         self.assertTrue(self.client.indices.exists("prod_index"))
-        self.assertEquals(
+        self.assertEqual(
             50, self.client.count(index="prod_index", q="type:answers")["count"]
         )
 
-        self.assertEquals(
+        self.assertEqual(
             {"answer": 42, "correct": True, "type": "answers"},
             self.client.get(index="prod_index", id=42)["_source"],
         )
@@ -551,22 +767,21 @@ class TestReindex(ElasticsearchTestCase):
         self.client.indices.refresh()
 
         self.assertTrue(self.client.indices.exists("prod_index"))
-        self.assertEquals(
+        self.assertEqual(
             50, self.client.count(index="prod_index", q="type:questions")["count"]
         )
-        self.assertEquals(
+        self.assertEqual(
             50, self.client.count(index="prod_index", q="type:answers")["count"]
         )
 
-        self.assertEquals(
+        self.assertEqual(
             {"answer": 42, "correct": True, "type": "answers"},
             self.client.get(index="prod_index", id=42)["_source"],
         )
 
 
 class TestParentChildReindex(ElasticsearchTestCase):
-    def setUp(self):
-        super(TestParentChildReindex, self).setUp()
+    def setup_method(self, _):
         body = {
             "settings": {"number_of_shards": 1, "number_of_replicas": 0},
             "mappings": {
@@ -596,7 +811,7 @@ class TestParentChildReindex(ElasticsearchTestCase):
         helpers.reindex(self.client, "test-index", "real-index")
 
         q = self.client.get(index="real-index", id=42)
-        self.assertEquals(
+        self.assertEqual(
             {
                 "_id": "42",
                 "_index": "real-index",
@@ -610,7 +825,7 @@ class TestParentChildReindex(ElasticsearchTestCase):
             q,
         )
         q = self.client.get(index="test-index", id=47, routing=42)
-        self.assertEquals(
+        self.assertEqual(
             {
                 "_routing": "42",
                 "_id": "47",
@@ -627,3 +842,60 @@ class TestParentChildReindex(ElasticsearchTestCase):
             },
             q,
         )
+
+
+@pytest.fixture(scope="function")
+def reindex_data_stream_setup(sync_client):
+    dt = datetime.now(tz=tz.UTC)
+    bulk = []
+    for x in range(100):
+        bulk.append({"index": {"_index": "test_index_stream", "_id": x}})
+        bulk.append(
+            {
+                "answer": x,
+                "correct": x == 42,
+                "type": "answers" if x % 2 == 0 else "questions",
+                "@timestamp": (dt - timedelta(days=x)).isoformat(),
+            }
+        )
+    sync_client.bulk(bulk, refresh=True)
+    sync_client.indices.put_index_template(
+        name="my-index-template",
+        body={
+            "index_patterns": ["py-*-*"],
+            "data_stream": {},
+        },
+    )
+    sync_client.indices.create_data_stream(name="py-test-stream")
+    sync_client.indices.refresh()
+
+
+class TestDataStreamReindex(object):
+    @pytest.mark.usefixtures("reindex_data_stream_setup")
+    @pytest.mark.parametrize("op_type", [None, "create"])
+    def test_reindex_index_datastream(self, op_type, sync_client):
+        helpers.reindex(
+            sync_client,
+            source_index="test_index_stream",
+            target_index="py-test-stream",
+            query={"query": {"bool": {"filter": {"term": {"type": "answers"}}}}},
+            op_type=op_type,
+        )
+        sync_client.indices.refresh()
+        assert sync_client.indices.exists(index="py-test-stream")
+        assert (
+            50 == sync_client.count(index="py-test-stream", q="type:answers")["count"]
+        )
+
+    @pytest.mark.usefixtures("reindex_data_stream_setup")
+    def test_reindex_index_datastream_op_type_index(self, sync_client):
+        with pytest.raises(
+            ValueError, match="Data streams must have 'op_type' set to 'create'"
+        ):
+            helpers.reindex(
+                sync_client,
+                source_index="test_index_stream",
+                target_index="py-test-stream",
+                query={"query": {"bool": {"filter": {"term": {"type": "answers"}}}}},
+                op_type="_index",
+            )
